/*
 *  Xen base ring buffer
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 * Copyright (C) 2016 EPAM Systems Inc.
 */

#ifndef XENBE_RINGBUFFERBASE_HPP_
#define XENBE_RINGBUFFERBASE_HPP_

#include <mutex>

extern "C" {
#include <xenctrl.h>
#include <xen/io/ring.h>
}

#include "XenEvtchn.hpp"
#include "Exception.hpp"
#include "XenGnttab.hpp"
#include "Log.hpp"

namespace XenBackend {

/***************************************************************************//**
 * Exception generated by RingBufferBase.
 * @ingroup backend
 ******************************************************************************/
class RingBufferException : public Exception
{
	using Exception::Exception;
};

/***************************************************************************//**
 * Interface to implement custom ring buffer.
 * @ingroup backend
 ******************************************************************************/
class RingBufferBase
{
public:

	/**
	 * @param domId frontend domain id
	 * @param port  event channel port number
	 * @param ref   grant table reference
	 */
	RingBufferBase(domid_t domId, evtchn_port_t port, grant_ref_t ref);
	virtual ~RingBufferBase();

	/**
	 * Starts ring buffer handling.
	 */
	void start();

	/**
	 * Stops ring buffer handling.
	 */
	void stop();

	/**
	 * Returns event channel port.
	 */
	evtchn_port_t getPort() const { return mPort; }

	/**
	 * Returns grant table reference.
	 */
	grant_ref_t getRef() const { return mRef; }

	/**
	 * Sets error callback
	 * @param errorCallback error callback
	 */
	void setErrorCallback(ErrorCallback errorCallback);

protected:

	/**
	 * Is called when the notification from associated event channel is
	 * received.
	 * This method should be reimplemented in derived classes.
	 */
	virtual void onReceiveIndication() = 0;

	/**
	 * Event channel.
	 */
	XenEvtchn mEventChannel;

	/**
	 * Grant table buffer.
	 */
	XenGnttabBuffer mBuffer;

	Log mLog;

private:

	evtchn_port_t mPort;
	grant_ref_t mRef;

	void onIndication();
};

/***************************************************************************//**
 * Base class to create the custom input ring buffer (for handling requests
 * from the frontend).
 * RingBufferInBase is a template with arguments taken from PV driver protocol.
 * The arguments of the template are structures defined with Xen
 * DEFINE_RING_TYPES() macro from ring.h. Also the in ring buffer takes a remote
 * event channel number and a grant reference on which the ring buffer is
 * mapped. Xen event channel is used to notify the backend that a new request is
 * available in the ring buffer. When a new request is received,
 * processRequest() method is called. To send the response, the client should
 * call sendResponse() method.
 *
 * In order to create the in ring buffer the client should implement a class
 * inherited from RingBufferInBase and override processRequest() method.
 *
 * @snippet ExampleBackend.hpp ExampleInRingBuffer
 *
 * processRequest():
 *
 * @snippet ExampleBackend.cpp processRequest
 *
 * @ingroup backend
 ******************************************************************************/
template<typename Ring, typename Page, typename Req, typename Rsp>
class RingBufferInBase : public RingBufferBase
{
public:

	/**
	 * @param[in] domId    frontend domain id
	 * @param[in] port     event channel port number
	 * @param[in] ref      ring buffer ref number
	 * @param[in] size ring buffer size
	 */
	RingBufferInBase(domid_t domId, evtchn_port_t port,
					 grant_ref_t ref, int size = XC_PAGE_SIZE) :
		RingBufferBase(domId, port, ref)
	{
		BACK_RING_INIT(&mRing, static_cast<Page*>(mBuffer.get()), size);
	}

protected:

	/**
	 * Processes frontend requests.
	 * This function is called when the request from the frontend is received
	 * and should be implemented in a derived class.
	 * @param req request
	 */
	virtual void processRequest(const Req& req) = 0;

	/**
	 * Sends the response to the frontend
	 * @param rsp response
	 */
	void sendResponse(const Rsp& rsp)
	{
		bool notify = false;

		*RING_GET_RESPONSE(&mRing, mRing.rsp_prod_pvt) = rsp;

		mRing.rsp_prod_pvt++;

		RING_PUSH_RESPONSES_AND_CHECK_NOTIFY(&mRing, notify);

		if (notify)
		{
			mEventChannel.notify();
		}
	}

private:

	Ring mRing;

	void onReceiveIndication()
	{
		int numPendingRequests = 0;

		do {
			Req req;

			auto rc = mRing.req_cons;
			auto rp = mRing.sring->req_prod;

			xen_rmb();

			if (RING_REQUEST_PROD_OVERFLOW(&mRing, rp))
			{
				throw RingBufferException("Ring buffer producer overflow", EIO);
			}

			while (rc != rp)
			{

				if (RING_REQUEST_CONS_OVERFLOW(&mRing, rc))
				{
					throw RingBufferException("Ring buffer consumer overflow", EIO);
				}

				req = *RING_GET_REQUEST(&mRing, rc);

				mRing.req_cons = ++rc;

				xen_mb();

				processRequest(req);
			}

			RING_FINAL_CHECK_FOR_REQUESTS(&mRing, numPendingRequests);
		}
		while (numPendingRequests);
	}
};

/***************************************************************************//**
 * Base class to create the custom output ring buffer (for sending events to
 * the frontend).
 * RingBufferOutBase is a template with arguments taken from PV driver protocol.
 * The arguments of the template are structures described the ring buffer and
 * events to be sent to the frontend. Also the out ring buffer takes a remote
 * event channel number, a grant reference on which the ring buffer is
 * mapped and offset where the events are stored in the ring buffer. Xen event
 * channel is used to notify the frontend that a new event is available in
 * the ring buffer.
 *
 * The out ring buffer can be instantiated directly, no need to implement a
 * class inherited from RingBufferOutBase.
 *
 * @snippet ExampleBackend.hpp ExampleOutRingBuffer
 *
 *The client should call sendEvent() method to send an event to the frontend:
 *
 * @snippet ExampleBackend.cpp onSomeEvent
 *
 * @ingroup backend
 ******************************************************************************/
template<typename Page,typename Event>
class RingBufferOutBase : public RingBufferBase
{
public:

	/**
	 * @param[in] domId    frontend domain id
	 * @param[in] port     event channel port number
	 * @param[in] ref      ring buffer ref number
	 * @param[in] offset   start of the ring buffer inside mapped page
	 * @param[in] size     size of the ring buffer
	 */
	RingBufferOutBase(domid_t domId, evtchn_port_t port, grant_ref_t ref,
					  int offset, size_t size) :
		RingBufferBase(domId, port, ref),
		mPage(static_cast<Page*>(mBuffer.get())),
		mEventBuffer(reinterpret_cast<Event*>(
				static_cast<uint8_t*>(mBuffer.get()) + offset)),
		mNumEvents(size/sizeof(Event))
	{
		mPage->in_prod = mPage->in_cons;

		xen_wmb();
	}

	/**
	 * Sends the event to the frontend
	 * @param event event to the frontend
	 */
	void sendEvent(const Event& event)
	{
		std::lock_guard<std::mutex> lock(mMutex);

		if (static_cast<int>(mPage->in_prod - mPage->in_cons) >= mNumEvents)
		{
			LOG(mLog, WARNING) << "Ring buffer overflow, port: " << getPort()
							   <<", prod: " << mPage->in_prod
							   << ", cons: " << mPage->in_cons;

			return;
		}

		LOG(mLog, DEBUG) << "Send event, port: " << getPort()
						 <<", prod: " << mPage->in_prod
						 << ", cons: " << mPage->in_cons
						 << ", num events: " << mNumEvents;

		mEventBuffer[mPage->in_prod % mNumEvents] = event;

		mPage->in_prod++;

		xen_wmb();

		mEventChannel.notify();
	}

protected:

	void onReceiveIndication() {}

private:

	Page* mPage;
	Event* mEventBuffer;
	int mNumEvents;

	std::mutex mMutex;
};

typedef std::shared_ptr<RingBufferBase> RingBufferPtr;

}

#endif /* XENBE_RINGBUFFERBASE_HPP_ */
